@@ -11,12 +11,14 @@ use chrono::offset::{Offset as ChronoOffset};
 use chrono::offset::TimeZone;
 use chrono::{Local as LocalTime};
 
+use std::clone::Clone;
+use std::cmp;
 use std::collections::HashMap;
 use std::default::Default;
 use std::mem;
 use std::slice;
 
-use crate::{PE, Error};
+use crate::{PE, PEType, PETranslation, Error};
 use crate::types::*;
 
 pub const DOS_SIGNATURE: u16    = 0x5A4D;
@@ -76,6 +78,55 @@ impl Default for ImageDOSHeader {
         }
     }
 }
+impl Clone for ImageDOSHeader {
+    fn clone(&self) -> Self {
+        // I have no idea why this is usafe but something similar below isn't.
+        unsafe {
+            Self {
+                e_magic: self.e_magic,
+                e_cblp: self.e_cblp,
+                e_cp: self.e_cp,
+                e_crlc: self.e_crlc,
+                e_cparhdr: self.e_cparhdr,
+                e_minalloc: self.e_minalloc,
+                e_maxalloc: self.e_maxalloc,
+                e_ss: self.e_ss,
+                e_sp: self.e_sp,
+                e_csum: self.e_csum,
+                e_ip: self.e_ip,
+                e_cs: self.e_cs,
+                e_lfarlc: self.e_lfarlc,
+                e_ovno: self.e_ovno,
+                e_res: self.e_res.clone(),
+                e_oemid: self.e_oemid,
+                e_oeminfo: self.e_oeminfo,
+                e_res2: self.e_res2.clone(),
+                e_lfanew: self.e_lfanew,
+            }
+        }
+    }
+    fn clone_from(&mut self, source: &Self) {
+        self.e_magic = source.e_magic;
+        self.e_cblp = source.e_cblp;
+        self.e_cp = source.e_cp;
+        self.e_crlc = source.e_crlc;
+        self.e_cparhdr = source.e_cparhdr;
+        self.e_minalloc = source.e_minalloc;
+        self.e_maxalloc = source.e_maxalloc;
+        self.e_ss = source.e_ss;
+        self.e_sp = source.e_sp;
+        self.e_csum = source.e_csum;
+        self.e_ip = source.e_ip;
+        self.e_cs = source.e_cs;
+        self.e_lfarlc = source.e_lfarlc;
+        self.e_ovno = source.e_ovno;
+        unsafe { self.e_res = source.e_res.clone(); }
+        self.e_oemid = source.e_oemid;
+        self.e_oeminfo = source.e_oeminfo;
+        unsafe { self.e_res2 = source.e_res2.clone(); }
+        self.e_lfanew = source.e_lfanew;
+    }
+}
 
 #[derive(Copy, Clone, Eq, PartialEq, Debug)]
 pub enum ImageFileMachine {
@@ -174,6 +225,28 @@ impl Default for ImageFileHeader {
         }
     }
 }
+impl Clone for ImageFileHeader {
+    fn clone(&self) -> Self {
+        Self {
+            machine: self.machine,
+            number_of_sections: self.number_of_sections,
+            time_date_stamp: self.time_date_stamp,
+            pointer_to_symbol_table: self.pointer_to_symbol_table,
+            number_of_symbols: self.number_of_symbols,
+            size_of_optional_header: self.size_of_optional_header,
+            characteristics: self.characteristics,
+        }
+    }
+    fn clone_from(&mut self, source: &Self) {
+        self.machine = source.machine;
+        self.number_of_sections = source.number_of_sections;
+        self.time_date_stamp = source.time_date_stamp;
+        self.pointer_to_symbol_table = source.pointer_to_symbol_table;
+        self.number_of_symbols = source.number_of_symbols;
+        self.size_of_optional_header = source.size_of_optional_header;
+        self.characteristics = source.characteristics;
+    }
+}
 
 #[derive(Copy, Clone, Eq, PartialEq, Debug)]
 pub enum ImageSubsystem {
@@ -247,7 +320,6 @@ pub struct ImageOptionalHeader32 {
     pub size_of_heap_commit: u32,
     pub loader_flags: u32,
     pub number_of_rva_and_sizes: u32,
-    pub data_directory: [ImageDataDirectory; 16],
 }
 impl Default for ImageOptionalHeader32 {
     fn default() -> Self {
@@ -282,29 +354,78 @@ impl Default for ImageOptionalHeader32 {
             size_of_heap_commit: 0x1000,
             loader_flags: 0,
             number_of_rva_and_sizes: 0x10,
-            /* I really don't want to give ImageDataDirectory the copy trait,
-               so this is ugly copypasta */
-            data_directory: [
-                ImageDataDirectory { virtual_address: RVA(0), size: 0 },
-                ImageDataDirectory { virtual_address: RVA(0), size: 0 },
-                ImageDataDirectory { virtual_address: RVA(0), size: 0 },
-                ImageDataDirectory { virtual_address: RVA(0), size: 0 },
-                ImageDataDirectory { virtual_address: RVA(0), size: 0 },
-                ImageDataDirectory { virtual_address: RVA(0), size: 0 },
-                ImageDataDirectory { virtual_address: RVA(0), size: 0 },
-                ImageDataDirectory { virtual_address: RVA(0), size: 0 },
-                ImageDataDirectory { virtual_address: RVA(0), size: 0 },
-                ImageDataDirectory { virtual_address: RVA(0), size: 0 },
-                ImageDataDirectory { virtual_address: RVA(0), size: 0 },
-                ImageDataDirectory { virtual_address: RVA(0), size: 0 },
-                ImageDataDirectory { virtual_address: RVA(0), size: 0 },
-                ImageDataDirectory { virtual_address: RVA(0), size: 0 },
-                ImageDataDirectory { virtual_address: RVA(0), size: 0 },
-                ImageDataDirectory { virtual_address: RVA(0), size: 0 },
-            ],
         }
     }
 }
+impl Clone for ImageOptionalHeader32 {
+    fn clone(&self) -> Self {
+        Self {
+            magic: self.magic,
+            major_linker_version: self.major_linker_version,
+            minor_linker_version: self.minor_linker_version,
+            size_of_code: self.size_of_code,
+            size_of_initialized_data: self.size_of_initialized_data,
+            size_of_uninitialized_data: self.size_of_uninitialized_data,
+            address_of_entry_point: self.address_of_entry_point,
+            base_of_code: self.base_of_code,
+            base_of_data: self.base_of_data,
+            image_base: self.image_base,
+            section_alignment: self.section_alignment,
+            file_alignment: self.file_alignment,
+            major_operating_system_version: self.major_operating_system_version,
+            minor_operating_system_version: self.minor_operating_system_version,
+            major_image_version: self.major_image_version,
+            minor_image_version: self.minor_image_version,
+            major_subsystem_version: self.major_subsystem_version,
+            minor_subsystem_version: self.minor_subsystem_version,
+            win32_version_value: self.win32_version_value,
+            size_of_image: self.size_of_image,
+            size_of_headers: self.size_of_headers,
+            checksum: self.checksum,
+            subsystem: self.subsystem,
+            dll_characteristics: self.dll_characteristics,
+            size_of_stack_reserve: self.size_of_stack_reserve,
+            size_of_stack_commit: self.size_of_stack_commit,
+            size_of_heap_reserve: self.size_of_heap_reserve,
+            size_of_heap_commit: self.size_of_heap_commit,
+            loader_flags: self.loader_flags,
+            number_of_rva_and_sizes: self.number_of_rva_and_sizes,
+        }
+    }
+
+    fn clone_from(&mut self, source: &Self) {
+        self.magic = source.magic;
+        self.major_linker_version = source.major_linker_version;
+        self.minor_linker_version = source.minor_linker_version;
+        self.size_of_code = source.size_of_code;
+        self.size_of_initialized_data = source.size_of_initialized_data;
+        self.size_of_uninitialized_data = source.size_of_uninitialized_data;
+        self.address_of_entry_point = source.address_of_entry_point;
+        self.base_of_code = source.base_of_code;
+        self.base_of_data = source.base_of_data;
+        self.image_base = source.image_base;
+        self.section_alignment = source.section_alignment;
+        self.file_alignment = source.file_alignment;
+        self.major_operating_system_version = source.major_operating_system_version;
+        self.minor_operating_system_version = source.minor_operating_system_version;
+        self.major_image_version = source.major_image_version;
+        self.minor_image_version = source.minor_image_version;
+        self.major_subsystem_version = source.major_subsystem_version;
+        self.minor_subsystem_version = source.minor_subsystem_version;
+        self.win32_version_value = source.win32_version_value;
+        self.size_of_image = source.size_of_image;
+        self.size_of_headers = source.size_of_headers;
+        self.checksum = source.checksum;
+        self.subsystem = source.subsystem;
+        self.dll_characteristics = source.dll_characteristics;
+        self.size_of_stack_reserve = source.size_of_stack_reserve;
+        self.size_of_stack_commit = source.size_of_stack_commit;
+        self.size_of_heap_reserve = source.size_of_heap_reserve;
+        self.size_of_heap_commit = source.size_of_heap_commit;
+        self.loader_flags = source.loader_flags;
+        self.number_of_rva_and_sizes = source.number_of_rva_and_sizes;
+    }
+}
 
 #[repr(packed)]
 pub struct ImageOptionalHeader64 {
@@ -337,7 +458,6 @@ pub struct ImageOptionalHeader64 {
     pub size_of_heap_commit: u64,
     pub loader_flags: u32,
     pub number_of_rva_and_sizes: u32,
-    pub data_directory: [ImageDataDirectory; 16],
 }
 impl Default for ImageOptionalHeader64 {
     fn default() -> Self {
@@ -371,27 +491,76 @@ impl Default for ImageOptionalHeader64 {
             size_of_heap_commit: 0x1000,
             loader_flags: 0,
             number_of_rva_and_sizes: 0x10,
-            data_directory: [
-                ImageDataDirectory { virtual_address: RVA(0), size: 0 },
-                ImageDataDirectory { virtual_address: RVA(0), size: 0 },
-                ImageDataDirectory { virtual_address: RVA(0), size: 0 },
-                ImageDataDirectory { virtual_address: RVA(0), size: 0 },
-                ImageDataDirectory { virtual_address: RVA(0), size: 0 },
-                ImageDataDirectory { virtual_address: RVA(0), size: 0 },
-                ImageDataDirectory { virtual_address: RVA(0), size: 0 },
-                ImageDataDirectory { virtual_address: RVA(0), size: 0 },
-                ImageDataDirectory { virtual_address: RVA(0), size: 0 },
-                ImageDataDirectory { virtual_address: RVA(0), size: 0 },
-                ImageDataDirectory { virtual_address: RVA(0), size: 0 },
-                ImageDataDirectory { virtual_address: RVA(0), size: 0 },
-                ImageDataDirectory { virtual_address: RVA(0), size: 0 },
-                ImageDataDirectory { virtual_address: RVA(0), size: 0 },
-                ImageDataDirectory { virtual_address: RVA(0), size: 0 },
-                ImageDataDirectory { virtual_address: RVA(0), size: 0 },
-            ],
         }
     }
 }
+impl Clone for ImageOptionalHeader64 {
+    fn clone(&self) -> Self {
+        Self {
+            magic: self.magic,
+            major_linker_version: self.major_linker_version,
+            minor_linker_version: self.minor_linker_version,
+            size_of_code: self.size_of_code,
+            size_of_initialized_data: self.size_of_initialized_data,
+            size_of_uninitialized_data: self.size_of_uninitialized_data,
+            address_of_entry_point: self.address_of_entry_point,
+            base_of_code: self.base_of_code,
+            image_base: self.image_base,
+            section_alignment: self.section_alignment,
+            file_alignment: self.file_alignment,
+            major_operating_system_version: self.major_operating_system_version,
+            minor_operating_system_version: self.minor_operating_system_version,
+            major_image_version: self.major_image_version,
+            minor_image_version: self.minor_image_version,
+            major_subsystem_version: self.major_subsystem_version,
+            minor_subsystem_version: self.minor_subsystem_version,
+            win32_version_value: self.win32_version_value,
+            size_of_image: self.size_of_image,
+            size_of_headers: self.size_of_headers,
+            checksum: self.checksum,
+            subsystem: self.subsystem,
+            dll_characteristics: self.dll_characteristics,
+            size_of_stack_reserve: self.size_of_stack_reserve,
+            size_of_stack_commit: self.size_of_stack_commit,
+            size_of_heap_reserve: self.size_of_heap_reserve,
+            size_of_heap_commit: self.size_of_heap_commit,
+            loader_flags: self.loader_flags,
+            number_of_rva_and_sizes: self.number_of_rva_and_sizes,
+        }
+    }
+
+    fn clone_from(&mut self, source: &Self) {
+        self.magic = source.magic;
+        self.major_linker_version = source.major_linker_version;
+        self.minor_linker_version = source.minor_linker_version;
+        self.size_of_code = source.size_of_code;
+        self.size_of_initialized_data = source.size_of_initialized_data;
+        self.size_of_uninitialized_data = source.size_of_uninitialized_data;
+        self.address_of_entry_point = source.address_of_entry_point;
+        self.base_of_code = source.base_of_code;
+        self.image_base = source.image_base;
+        self.section_alignment = source.section_alignment;
+        self.file_alignment = source.file_alignment;
+        self.major_operating_system_version = source.major_operating_system_version;
+        self.minor_operating_system_version = source.minor_operating_system_version;
+        self.major_image_version = source.major_image_version;
+        self.minor_image_version = source.minor_image_version;
+        self.major_subsystem_version = source.major_subsystem_version;
+        self.minor_subsystem_version = source.minor_subsystem_version;
+        self.win32_version_value = source.win32_version_value;
+        self.size_of_image = source.size_of_image;
+        self.size_of_headers = source.size_of_headers;
+        self.checksum = source.checksum;
+        self.subsystem = source.subsystem;
+        self.dll_characteristics = source.dll_characteristics;
+        self.size_of_stack_reserve = source.size_of_stack_reserve;
+        self.size_of_stack_commit = source.size_of_stack_commit;
+        self.size_of_heap_reserve = source.size_of_heap_reserve;
+        self.size_of_heap_commit = source.size_of_heap_commit;
+        self.loader_flags = source.loader_flags;
+        self.number_of_rva_and_sizes = source.number_of_rva_and_sizes;
+    }
+}
 
 #[repr(packed)]
 pub struct ImageNTHeaders32 {
@@ -408,6 +577,20 @@ impl Default for ImageNTHeaders32 {
         }
     }
 }
+impl Clone for ImageNTHeaders32 {
+    fn clone(&self) -> Self {
+        Self {
+            signature: self.signature,
+            file_header: self.file_header.clone(),
+            optional_header: self.optional_header.clone(),
+        }
+    }
+    fn clone_from(&mut self, source: &Self) {
+        self.signature = source.signature;
+        self.file_header = source.file_header.clone();
+        self.optional_header = source.optional_header.clone();
+    }
+}
 
 #[repr(packed)]
 pub struct ImageNTHeaders64 {
@@ -424,6 +607,20 @@ impl Default for ImageNTHeaders64 {
         }
     }
 }
+impl Clone for ImageNTHeaders64 {
+    fn clone(&self) -> Self {
+        Self {
+            signature: self.signature,
+            file_header: self.file_header.clone(),
+            optional_header: self.optional_header.clone(),
+        }
+    }
+    fn clone_from(&mut self, source: &Self) {
+        self.signature = source.signature;
+        self.file_header = source.file_header.clone();
+        self.optional_header = source.optional_header.clone();
+    }
+}
 
 bitflags! {
     /// A series of bitflags representing section characteristics.
@@ -491,6 +688,125 @@ pub struct ImageSectionHeader {
     pub number_of_linenumbers: u16,
     pub characteristics: SectionCharacteristics,
 }
+impl ImageSectionHeader {
+    /// Set the name of this section. The name will be truncated to eight bytes. If ```name``` is [`None`](Option::None), it
+    /// zeroes out the name field.
+    pub fn set_name(&mut self, name: Option<&str>) {
+        self.name.copy_from_slice((0..8)
+                                  .map(|_| CChar(0))
+                                  .collect::<Vec<CChar>>()
+                                  .as_slice());
+        
+        if name.is_none() {
+            return;
+        }
+
+        let new_name = name.unwrap();
+        let name_len = cmp::min(new_name.len(),8);
+        let name_vec = new_name
+            .as_bytes()[..name_len]
+            .iter()
+            .map(|&x| CChar(x))
+            .collect::<Vec<CChar>>();
+        let padding = (0..(8-name_len))
+            .map(|_| CChar(0))
+            .collect::<Vec<CChar>>();
+        
+        self.name.copy_from_slice(name_vec
+                                  .iter()
+                                  .chain(padding.iter())
+                                  .map(|&x| x)
+                                  .collect::<Vec<CChar>>()
+                                  .as_slice());
+    }
+
+    /// Check whether the given [`Offset`](Offset) is in this section.
+    pub fn has_offset(&self, offset: Offset) -> bool {
+        let start = self.pointer_to_raw_data;
+        let end = Offset(start.0 + self.size_of_raw_data);
+
+        start.0 <= offset.0 && offset.0 < end.0
+    }
+    /// Check whether the given [`RVA`](RVA) is in this section.
+    pub fn has_rva(&self, rva: RVA) -> bool {
+        let start = self.virtual_address;
+        let end = RVA(start.0 + self.virtual_size);
+
+        start.0 <= rva.0 && rva.0 < end.0
+    }
+
+    /// Get the offset to the data this section represents. This essentially performs the same task as
+    /// [`PE::translate`](PE::translate).
+    pub fn data_offset(&self, pe_type: PEType) -> Offset {
+        match pe_type {
+            PEType::Disk => self.pointer_to_raw_data,
+            PEType::Memory => Offset(self.virtual_address.0),
+        }
+    }
+
+    /// Get the size of this section.
+    pub fn data_size(&self, pe_type: PEType) -> usize {
+        match pe_type {
+            PEType::Disk => self.size_of_raw_data as usize,
+            PEType::Memory => self.virtual_size as usize,
+        }
+    }
+
+    /// Read a slice of the data this section represents.
+    pub fn read<'data>(&'data self, pe: &'data PE) -> Result<&'data [u8], Error> {
+        let offset = self.data_offset(pe.pe_type);
+        let size = self.data_size(pe.pe_type);
+
+        pe.buffer.read(offset, size)
+    }
+    /// Read a mutable slice of the data this section represents.
+    pub fn read_mut<'data>(&'data self, pe: &'data mut PE) -> Result<&'data mut [u8], Error> {
+        let offset = self.data_offset(pe.pe_type);
+        let size = self.data_size(pe.pe_type);
+
+        pe.buffer.read_mut(offset, size)
+    }
+    /// Write data to this section. It returns [`Error::BufferTooSmall`](Error::BufferTooSmall) if the data
+    /// overflows the section.
+    pub fn write(&self, pe: &mut PE, data: &[u8]) -> Result<(), Error> {
+        let offset = self.data_offset(pe.pe_type);
+        let size = self.data_size(pe.pe_type);
+
+        if data.len() > size {
+            return Err(Error::BufferTooSmall);
+        }
+
+        pe.buffer.write(offset, data)
+    }
+}
+impl Clone for ImageSectionHeader {
+    fn clone(&self) -> Self {
+        Self {
+            name: self.name.clone(),
+            virtual_size: self.virtual_size,
+            virtual_address: self.virtual_address,
+            size_of_raw_data: self.size_of_raw_data,
+            pointer_to_raw_data: self.pointer_to_raw_data,
+            pointer_to_relocations: self.pointer_to_relocations,
+            pointer_to_linenumbers: self.pointer_to_linenumbers,
+            number_of_relocations: self.number_of_relocations,
+            number_of_linenumbers: self.number_of_linenumbers,
+            characteristics: self.characteristics,
+        }
+    }
+    fn clone_from(&mut self, source: &Self) {
+        self.name = source.name.clone();
+        self.virtual_size = source.virtual_size;
+        self.virtual_address = source.virtual_address;
+        self.size_of_raw_data = source.size_of_raw_data;
+        self.pointer_to_raw_data = source.pointer_to_raw_data;
+        self.pointer_to_relocations = source.pointer_to_relocations;
+        self.pointer_to_linenumbers = source.pointer_to_linenumbers;
+        self.number_of_relocations = source.number_of_relocations;
+        self.number_of_linenumbers = source.number_of_linenumbers;
+        self.characteristics = source.characteristics;
+    }
+}
 
 #[derive(Copy, Clone, Eq, PartialEq, Debug)]
 pub enum ImageDirectoryEntry {
@@ -525,7 +841,7 @@ impl ImageDataDirectory {
             return Err(Error::InvalidRVA);
         }
         
-        let address = match self.virtual_address.as_offset(pe) {
+        let address = match pe.translate(PETranslation::Memory(self.virtual_address)) {
             Ok(a) => a,
             Err(e) => return Err(e),
         };
@@ -559,7 +875,7 @@ impl ImageDataDirectory {
             return Err(Error::InvalidRVA);
         }
         
-        let address = match self.virtual_address.as_offset(pe) {
+        let address = match pe.translate(PETranslation::Memory(self.virtual_address)) {
             Ok(a) => a,
             Err(e) => return Err(e),
         };
@@ -588,6 +904,19 @@ impl ImageDataDirectory {
         }
     }
 }
+impl Clone for ImageDataDirectory {
+    fn clone(&self) -> Self {
+        Self {
+            virtual_address: self.virtual_address,
+            size: self.size,
+        }
+    }
+
+    fn clone_from(&mut self, source: &Self) {
+        self.virtual_address = source.virtual_address;
+        self.size = source.size;
+    }
+}
 
 #[repr(packed)]
 pub struct ImageExportDirectory {
@@ -610,7 +939,7 @@ impl ImageExportDirectory {
             return Err(Error::InvalidRVA);
         }
         
-        match self.name.as_offset(pe) {
+        match pe.translate(PETranslation::Memory(self.name)) {
             Err(e) => return Err(e),
             Ok(a) => pe.buffer.get_cstring(a, false, None),
         }
@@ -621,7 +950,7 @@ impl ImageExportDirectory {
             return Err(Error::InvalidRVA);
         }
 
-        match self.name.as_offset(pe) {
+        match pe.translate(PETranslation::Memory(self.name)) {
             Err(e) => return Err(e),
             Ok(a) => pe.buffer.get_mut_cstring(a, false, None),
         }
@@ -634,7 +963,7 @@ impl ImageExportDirectory {
             return Err(Error::InvalidRVA);
         }
 
-        match self.address_of_functions.as_offset(pe) {
+        match pe.translate(PETranslation::Memory(self.address_of_functions)) {
             Err(e) => return Err(e),
             Ok(a) => pe.buffer.get_slice_ref::<Thunk32>(a, self.number_of_functions as usize),
         }
@@ -645,7 +974,7 @@ impl ImageExportDirectory {
             return Err(Error::InvalidRVA);
         }
 
-        match self.address_of_functions.as_offset(pe) {
+        match pe.translate(PETranslation::Memory(self.address_of_functions)) {
             Err(e) => return Err(e),
             Ok(a) => pe.buffer.get_mut_slice_ref::<Thunk32>(a, self.number_of_functions as usize),
         }
@@ -657,7 +986,7 @@ impl ImageExportDirectory {
             return Err(Error::InvalidRVA);
         }
 
-        match self.address_of_names.as_offset(pe) {
+        match pe.translate(PETranslation::Memory(self.address_of_names)) {
             Err(e) => return Err(e),
             Ok(a) => pe.buffer.get_slice_ref::<RVA>(a, self.number_of_names as usize),
         }
@@ -668,7 +997,7 @@ impl ImageExportDirectory {
             return Err(Error::InvalidRVA);
         }
 
-        match self.address_of_names.as_offset(pe) {
+        match pe.translate(PETranslation::Memory(self.address_of_names)) {
             Err(e) => return Err(e),
             Ok(a) => pe.buffer.get_mut_slice_ref::<RVA>(a, self.number_of_names as usize),
         }
@@ -680,7 +1009,7 @@ impl ImageExportDirectory {
             return Err(Error::InvalidRVA);
         }
 
-        match self.address_of_name_ordinals.as_offset(pe) {
+        match pe.translate(PETranslation::Memory(self.address_of_name_ordinals)) {
             Err(e) => return Err(e),
             Ok(a) => pe.buffer.get_slice_ref::<u16>(a, self.number_of_names as usize),
         }
@@ -691,7 +1020,7 @@ impl ImageExportDirectory {
             return Err(Error::InvalidRVA);
         }
 
-        match self.address_of_name_ordinals.as_offset(pe) {
+        match pe.translate(PETranslation::Memory(self.address_of_name_ordinals)) {
             Err(e) => return Err(e),
             Ok(a) => pe.buffer.get_mut_slice_ref::<u16>(a, self.number_of_names as usize),
         }
@@ -729,7 +1058,7 @@ impl ImageExportDirectory {
             let name_rva = names[index as usize];
             if name_rva.0 == 0 { continue; }
 
-            let name_offset = match name_rva.as_offset(pe) {
+            let name_offset = match pe.translate(PETranslation::Memory(name_rva)) {
                 Ok(o) => o,
                 Err(_) => continue, /* we continue instead of returning the error to be greedy with parsing */
             };
@@ -748,6 +1077,36 @@ impl ImageExportDirectory {
         Ok(result)
     }
 }
+impl Clone for ImageExportDirectory {
+    fn clone(&self) -> Self {
+        Self {
+            characteristics: self.characteristics,
+            time_date_stamp: self.time_date_stamp,
+            major_version: self.major_version,
+            minor_version: self.minor_version,
+            name: self.name,
+            base: self.base,
+            number_of_functions: self.number_of_functions,
+            number_of_names: self.number_of_names,
+            address_of_functions: self.address_of_functions,
+            address_of_names: self.address_of_names,
+            address_of_name_ordinals: self.address_of_name_ordinals,
+        }
+    }
+    fn clone_from(&mut self, source: &Self) {
+        self.characteristics = source.characteristics;
+        self.time_date_stamp = source.time_date_stamp;
+        self.major_version = source.major_version;
+        self.minor_version = source.minor_version;
+        self.name = source.name;
+        self.base = source.base;
+        self.number_of_functions = source.number_of_functions;
+        self.number_of_names = source.number_of_names;
+        self.address_of_functions = source.address_of_functions;
+        self.address_of_names = source.address_of_names;
+        self.address_of_name_ordinals = source.address_of_name_ordinals;
+    }
+}
 
 #[repr(packed)]
 pub struct ImageImportDescriptor {
@@ -769,7 +1128,7 @@ impl ImageImportDescriptor {
         };
         
         let mut thunks = 0usize;
-        let mut indexer = match rva.as_offset(pe) {
+        let mut indexer = match pe.translate(PETranslation::Memory(rva)) {
             Ok(i) => i,
             Err(e) => return Err(e),
         };
@@ -814,7 +1173,7 @@ impl ImageImportDescriptor {
             Err(e) => return Err(e),
         };
 
-        let offset = match rva.as_offset(pe) {
+        let offset = match pe.translate(PETranslation::Memory(rva)) {
             Ok(o) => o,
             Err(e) => return Err(e),
         };
@@ -845,7 +1204,7 @@ impl ImageImportDescriptor {
             Err(e) => return Err(e),
         };
 
-        let offset = match rva.as_offset(pe) {
+        let offset = match pe.translate(PETranslation::Memory(rva)) {
             Ok(o) => o,
             Err(e) => return Err(e),
         };
@@ -867,8 +1226,8 @@ impl ImageImportDescriptor {
         if dir.virtual_address.0 == 0 {
             return Err(Error::InvalidRVA);
         }
-        
-        let mut address = match dir.virtual_address.as_offset(pe) {
+
+        let mut address = match pe.translate(PETranslation::Memory(dir.virtual_address)) {
             Ok(a) => a,
             Err(e) => return Err(e),
         };
@@ -900,7 +1259,7 @@ impl ImageImportDescriptor {
             Err(e) => return Err(e),
         };
 
-        let offset = match dir.virtual_address.as_offset(pe) {
+        let offset = match pe.translate(PETranslation::Memory(dir.virtual_address)) {
             Ok(o) => o,
             Err(e) => return Err(e),
         };
@@ -916,7 +1275,7 @@ impl ImageImportDescriptor {
             Err(e) => return Err(e),
         };
 
-        let offset = match dir.virtual_address.as_offset(pe) {
+        let offset = match pe.translate(PETranslation::Memory(dir.virtual_address)) {
             Ok(o) => o,
             Err(e) => return Err(e),
         };
@@ -935,7 +1294,7 @@ impl ImageImportDescriptor {
 
     /// Get the name of the module represented by this import descriptor entry.
     pub fn get_name<'data>(&self, pe: &'data PE) -> Result<&'data [CChar], Error> {
-        let offset = match self.name.as_offset(pe) {
+        let offset = match pe.translate(PETranslation::Memory(self.name)) {
             Ok(o) => o,
             Err(e) => return Err(e),
         };
@@ -944,7 +1303,7 @@ impl ImageImportDescriptor {
     }
     /// Get the mutable name of the module represented by this import descriptor entry.
     pub fn get_mut_name<'data>(&self, pe: &'data mut PE) -> Result<&'data mut [CChar], Error> {
-        let offset = match self.name.as_offset(pe) {
+        let offset = match pe.translate(PETranslation::Memory(self.name)) {
             Ok(o) => o,
             Err(e) => return Err(e),
         };
@@ -1001,6 +1360,24 @@ impl ImageImportDescriptor {
         Ok(results)
     }
 }
+impl Clone for ImageImportDescriptor {
+    fn clone(&self) -> Self {
+        Self {
+            original_first_thunk: self.original_first_thunk,
+            time_date_stamp: self.time_date_stamp,
+            forwarder_chain: self.forwarder_chain,
+            name: self.name,
+            first_thunk: self.first_thunk,
+        }
+    }
+    fn clone_from(&mut self, source: &Self) {
+        self.original_first_thunk = source.original_first_thunk;
+        self.time_date_stamp = source.time_date_stamp;
+        self.forwarder_chain = source.forwarder_chain;
+        self.name = source.name;
+        self.first_thunk = source.first_thunk;
+    }
+}
 
 /// Represents an ```IMAGE_IMPORT_BY_NAME``` structure.
 ///
@@ -1016,7 +1393,7 @@ pub struct ImageImportByName<'data> {
 impl<'data> ImageImportByName<'data> {
     /// Get an ```ImageImportByName``` object at the given RVA.
     pub fn parse(pe: &'data PE, rva: RVA) -> Result<Self, Error> {
-        let offset = match rva.as_offset(pe) {
+        let offset = match pe.translate(PETranslation::Memory(rva)) {
             Ok(o) => o,
             Err(e) => return Err(e),
         };
@@ -1042,14 +1419,14 @@ pub struct ImageImportByNameMut<'data> {
 impl<'data> ImageImportByNameMut<'data> {
     /// Get a mutable ```ImageImportByName``` object at the given RVA.
     pub fn parse(pe: &'data mut PE, rva: RVA) -> Result<Self, Error> {
-        let mut offset = match rva.as_offset(pe) {
+        let mut offset = match pe.translate(PETranslation::Memory(rva)) {
             Ok(o) => o,
             Err(e) => return Err(e),
         };
 
-        let mut ptr = unsafe { pe.buffer.offset_to_mut_ptr(offset) };
-
         unsafe {
+            let mut ptr = pe.buffer.offset_to_mut_ptr(offset);
+
             let hint = &mut *(ptr as *mut u16);
             let u16_size = mem::size_of::<u16>();
 
@@ -1095,3 +1472,15 @@ pub struct ImageBaseRelocation {
     pub virtual_address: RVA,
     pub size_of_block: u32,
 }
+impl Clone for ImageBaseRelocation {
+    fn clone(&self) -> Self {
+        Self {
+            virtual_address: self.virtual_address,
+            size_of_block: self.size_of_block,
+        }
+    }
+    fn clone_from(&mut self, source: &Self) {
+        self.virtual_address = source.virtual_address;
+        self.size_of_block = source.size_of_block;
+    }
+}
