@@ -32,7 +32,7 @@ mod tests;
 
 use std::convert::AsRef;
 use std::io::{Error as IoError};
-use std::mem::size_of;
+use std::mem;
 use std::path::Path;
 
 use crate::buffer::Buffer;
@@ -66,11 +66,57 @@ pub enum Error {
     UnsupportedDirectory,
     /// The relocation entry is invalid.
     InvalidRelocation,
+    /// The provided directory is not available.
+    BadDirectory,
+}
+
+/// An enum to tag the PE file with what its memory map looks like.
+///
+/// When a PE is loaded, it's ultimately parsed before being placed into memory. This
+/// means the image in memory is different from the disk. This is why, for example,
+/// RVAs and Offsets differ in type-- the RVA represents the offset to the image in
+/// *memory*, whereas the Offset represents the offset to the image on *disk*. This enum
+/// is necessary to maintain a simple translation layer for basic address operations.
+#[derive(Copy, Clone, Eq, PartialEq, Debug)]
+pub enum PEType {
+    Disk,
+    Memory,
+}
+
+/// An enum to translate between RVA and Offset addresses.
+///
+/// This typically never gets exposed beyond the [translate](PE::translate) function. See [PEType](PEType)
+/// for an explanation of why this is here.
+pub enum PETranslation {
+    Disk(Offset),
+    Memory(RVA),
+}
+impl Address for PETranslation {
+    fn as_offset(&self, pe: &PE) -> Result<Offset, Error> {
+        match self {
+            Self::Disk(o) => Ok(*o),
+            Self::Memory(r) => r.as_offset(pe),
+        }
+    }
+    fn as_rva(&self, pe: &PE) -> Result<RVA, Error> {
+        match self {
+            Self::Disk(o) => o.as_rva(pe),
+            Self::Memory(r) => Ok(*r),
+        }
+    }
+    fn as_va(&self, pe: &PE) -> Result<VA, Error> {
+        match self {
+            Self::Disk(o) => o.as_va(pe),
+            Self::Memory(r) => r.as_va(pe),
+        }
+    }
 }
 
 /// Represents a PE file.
 #[derive(Clone, Eq, PartialEq, Debug)]
 pub struct PE {
+    /// The type of buffer the PE file is expecting. See [PEType](PEType) for an explanation.
+    pub pe_type: PEType,
     /// The buffer that holds the data. Various operations such as getting
     /// references to objects in the data can be found in the buffer object.
     pub buffer: Buffer,
@@ -81,15 +127,17 @@ pub struct PE {
 impl PE {
     /// Generates a new, blank PE file. Typically only useful for constructing
     /// new PE files.
-    pub fn new(size: Option<usize>) -> Self {
+    pub fn new(size: Option<usize>, pe_type: PEType) -> Self {
         Self {
+            pe_type: pe_type,
             buffer: Buffer::new(size),
             filename: None,
         }
     }
-    /// Generates a new PE object from a slice of data.
-    pub fn from_data(data: &[u8]) -> Self {
+    /// Generates a new PE object from a slice of data, marking it as a memory-resident image.
+    pub fn from_data(data: &[u8], pe_type: PEType) -> Self {
         Self {
+            pe_type: pe_type,
             buffer: Buffer::from_data(data),
             filename: None,
         }
@@ -97,10 +145,51 @@ impl PE {
     /// Generates a new PE object from a file on disk.
     pub fn from_file<P: AsRef<Path>>(filename: P) -> Result<Self, IoError> {
         match Buffer::from_file(&filename) {
-            Ok(buffer) => Ok(Self { buffer: buffer, filename: Some(String::from(filename.as_ref().to_str().unwrap())) }),
+            Ok(buffer) => Ok(
+                Self {
+                    pe_type: PEType::Disk,
+                    buffer: buffer,
+                    filename: Some(String::from(filename.as_ref().to_str().unwrap()))
+                }
+            ),
             Err(e) => Err(e),
         }
     }
+    /// Generates a new PE object from a file on disk, marking it as a memory dump (i.e., sets ```pe_type``` to [PEType::Memory](PEType::Memory).
+    pub fn from_memory_dump<P: AsRef<Path>>(filename: P) -> Result<Self, IoError> {
+        match Buffer::from_file(&filename) {
+            Ok(buffer) => Ok(
+                Self {
+                    pe_type: PEType::Memory,
+                    buffer: buffer,
+                    filename: Some(String::from(filename.as_ref().to_str().unwrap()))
+                }
+            ),
+            Err(e) => Err(e),
+        }
+    }
+
+    /// Translate an address into a buffer offset relevant to the image type.
+    ///
+    /// This differs from [rva_to_offset](PE::rva_to_offset) because it does not directly rely on the section table.
+    /// Rather, if the image is a memory image, it treats [RVA](RVA)s as offsets, because that's what they are in memory.
+    /// Otherwise, it converts the [RVA](RVA) into an offset via the section table. The reverse goes for if
+    /// the PE image is a disk image and an [Offset](Offset) is provided.
+    pub fn translate(&self, addr: PETranslation) -> Result<Offset, Error> {
+        match self.pe_type {
+            PEType::Disk => match addr {
+                PETranslation::Disk(o) => Ok(o),
+                PETranslation::Memory(r) => r.as_offset(self),
+            }
+            PEType::Memory => match addr {
+                PETranslation::Disk(o) => match o.as_rva(self) {
+                    Ok(rva) => Ok(Offset(rva.0)),
+                    Err(e) => Err(e),
+                },
+                PETranslation::Memory(r) => Ok(Offset(r.0)),
+            }
+        }
+    }
 
     /// Get the DOS header without verifying its contents.
     pub fn get_dos_header(&self) -> Result<&ImageDOSHeader, Error> {
@@ -370,6 +459,99 @@ impl PE {
         }
     }
 
+    /// Get the entrypoint of this PE file.
+    pub fn get_entrypoint(&self) -> Result<RVA, Error> {
+        let nt_headers = match self.get_valid_nt_headers() {
+            Ok(h) => h,
+            Err(e) => return Err(e),
+        };
+
+        match nt_headers {
+            NTHeaders::NTHeaders32(h32) => Ok(h32.optional_header.address_of_entry_point),
+            NTHeaders::NTHeaders64(h64) => Ok(h64.optional_header.address_of_entry_point),
+        }
+    }
+
+    /// Get the offset to the data directory within the PE file.
+    pub fn get_data_directory_offset(&self) -> Result<Offset, Error> {
+        let e_lfanew = match self.e_lfanew() {
+            Ok(o) => o,
+            Err(e) => return Err(e),
+        };
+        
+        let nt_header = match self.get_valid_nt_headers() {
+            Ok(h) => h,
+            Err(e) => return Err(e),
+        };
+
+        let header_size = match nt_header {
+            NTHeaders::NTHeaders32(_) => mem::size_of::<ImageNTHeaders32>(),
+            NTHeaders::NTHeaders64(_) => mem::size_of::<ImageNTHeaders64>(),
+        };
+
+        let offset = Offset(e_lfanew.0 + (header_size as u32));
+
+        if !self.validate_offset(offset) {
+            return Err(Error::BufferTooSmall);
+        }
+
+        Ok(offset)
+    }
+    /// Get the size of the data directory. Rounds down ```number_of_rva_and_sizes``` to 16, which is what
+    /// the Windows loader does.
+    pub fn get_data_directory_size(&self) -> Result<usize, Error> {
+        let nt_header = match self.get_valid_nt_headers() {
+            Ok(h) => h,
+            Err(e) => return Err(e),
+        };
+
+        let sizes = match nt_header {
+            NTHeaders::NTHeaders32(h32) => h32.optional_header.number_of_rva_and_sizes,
+            NTHeaders::NTHeaders64(h64) => h64.optional_header.number_of_rva_and_sizes,
+        };
+
+        // data directory gets rounded down if greater than 16
+        if sizes > 16 {
+            Ok(16 as usize)
+        }
+        else {
+            Ok(sizes as usize)
+        }
+    }
+    /// Get the data directory table.
+    ///
+    /// Normally one would expect this to be a part of [ImageOptionalHeader](ImageOptionalHeader32), but
+    /// [ImageOptionalHeader::number_of_rva_and_sizes](ImageOptionalHeader32::number_of_rva_and_sizes) controls
+    /// the size of the array, so we can't exactly stick it in the optional header, because that would
+    /// produce a variable-sized structure.
+    pub fn get_data_directory_table(&self) -> Result<&[ImageDataDirectory], Error> {
+        let offset = match self.get_data_directory_offset() {
+            Ok(o) => o,
+            Err(e) => return Err(e),
+        };
+
+        let size = match self.get_data_directory_size() {
+            Ok(s) => s,
+            Err(e) => return Err(e),
+        };
+
+        self.buffer.get_slice_ref::<ImageDataDirectory>(offset, size)
+    }
+    /// Get a mutable data directory table.
+    pub fn get_mut_data_directory_table(&mut self) -> Result<&mut [ImageDataDirectory], Error> {
+        let offset = match self.get_data_directory_offset() {
+            Ok(o) => o,
+            Err(e) => return Err(e),
+        };
+
+        let size = match self.get_data_directory_size() {
+            Ok(s) => s,
+            Err(e) => return Err(e),
+        };
+
+        self.buffer.get_mut_slice_ref::<ImageDataDirectory>(offset, size)
+    }
+
     /// Get the offset to the section table within the PE file.
     pub fn get_section_table_offset(&self) -> Result<Offset, Error> {
         let e_lfanew = match self.e_lfanew() {
@@ -389,8 +571,8 @@ impl PE {
 
         let Offset(mut offset) = e_lfanew;
 
-        offset += size_of::<u32>() as u32;
-        offset += size_of::<ImageFileHeader>() as u32;
+        offset += mem::size_of::<u32>() as u32;
+        offset += mem::size_of::<ImageFileHeader>() as u32;
         offset += size_of_optional as u32;
 
         if !self.validate_offset(Offset(offset)) {
@@ -447,10 +629,7 @@ impl PE {
         };
 
         for section in section_table {
-            let base = section.pointer_to_raw_data.0;
-            let end = base+section.size_of_raw_data;
-
-            if base <= offset.0 && offset.0 < end {
+            if section.has_offset(offset) {
                 return Ok(section);
             }
         }
@@ -467,10 +646,7 @@ impl PE {
         };
 
         for section in section_table {
-            let base = section.pointer_to_raw_data.0;
-            let end = base+section.size_of_raw_data;
-
-            if base <= offset.0 && offset.0 < end {
+            if section.has_offset(offset) {
                 return Ok(section);
             }
         }
@@ -487,10 +663,7 @@ impl PE {
         };
 
         for section in section_table {
-            let base = section.virtual_address.0;
-            let end = base+section.virtual_size;
-
-            if base <= rva.0 && rva.0 < end {
+            if section.has_rva(rva) {
                 return Ok(section);
             }
         }
@@ -507,10 +680,7 @@ impl PE {
         };
 
         for section in section_table {
-            let base = section.virtual_address.0;
-            let end = base+section.virtual_size;
-
-            if base <= rva.0 && rva.0 < end {
+            if section.has_rva(rva) {
                 return Ok(section);
             }
         }
@@ -598,7 +768,7 @@ impl PE {
     }
 
     /// Convert an offset to an RVA address. Produces [Error::InvalidRVA](Error::InvalidRVA) if the produced
-    /// RVA is invalid.
+    /// RVA is invalid or if the section it was transposed from no longer contains it.
     pub fn offset_to_rva(&self, offset: Offset) -> Result<RVA, Error> {
         let section = match self.get_section_by_offset(offset) {
             Ok(s) => s,
@@ -621,7 +791,7 @@ impl PE {
 
         let final_rva = RVA(rva);
 
-        if !self.validate_rva(final_rva) {
+        if !self.validate_rva(final_rva) || !section.has_rva(final_rva) {
             return Err(Error::InvalidRVA);
         }
 
@@ -638,7 +808,7 @@ impl PE {
     }
 
     /// Convert an RVA to an offset address. Produces a [Error::InvalidOffset](Error::InvalidOffset) error if
-    /// the produced offset is invalid.
+    /// the produced offset is invalid or if the section it was transposed from no longer contains it.
     pub fn rva_to_offset(&self, rva: RVA) -> Result<Offset, Error> {
         let section = match self.get_section_by_rva(rva) {
             Ok(s) => s,
@@ -661,7 +831,7 @@ impl PE {
 
         let final_offset = Offset(offset);
 
-        if !self.validate_offset(final_offset) {
+        if !self.validate_offset(final_offset) || !section.has_offset(final_offset) {
             return Err(Error::InvalidOffset);
         }
 
@@ -720,24 +890,34 @@ impl PE {
     }
 
     /// Get the data directory reference represented by the [ImageDirectoryEntry](headers::ImageDirectoryEntry) enum.
+    /// Returns [Error::BadDirectory](Error::BadDirectory) if the given directory is inaccessible due to the directory
+    /// size.
     pub fn get_data_directory(&self, dir: ImageDirectoryEntry) -> Result<&ImageDataDirectory, Error> {
-        match self.get_valid_nt_headers() {
+        let directory_table = match self.get_data_directory_table() {
+            Ok(d) => d,
             Err(e) => return Err(e),
-            Ok(h) => match h {
-                NTHeaders::NTHeaders32(h32) => Ok(&h32.optional_header.data_directory[dir as usize]),
-                NTHeaders::NTHeaders64(h64) => Ok(&h64.optional_header.data_directory[dir as usize]),
-            }
+        };
+        let index = dir as usize;
+
+        if index >= directory_table.len() {
+            return Err(Error::BadDirectory);
         }
+
+        Ok(&directory_table[index])
     }
     /// Get the mutable data directory reference represented by the [ImageDirectoryEntry](headers::ImageDirectoryEntry) enum.
     pub fn get_mut_data_directory(&mut self, dir: ImageDirectoryEntry) -> Result<&mut ImageDataDirectory, Error> {
-        match self.get_valid_mut_nt_headers() {
+        let directory_table = match self.get_mut_data_directory_table() {
+            Ok(d) => d,
             Err(e) => return Err(e),
-            Ok(h) => match h {
-                NTHeadersMut::NTHeaders32(h32) => Ok(&mut h32.optional_header.data_directory[dir as usize]),
-                NTHeadersMut::NTHeaders64(h64) => Ok(&mut h64.optional_header.data_directory[dir as usize]),
-            }
+        };
+        let index = dir as usize;
+
+        if index >= directory_table.len() {
+            return Err(Error::BadDirectory);
         }
+
+        Ok(&mut directory_table[index])
     }
     
     /// Resolve the data directory represented by the [ImageDirectoryEntry](headers::ImageDirectoryEntry) enum. This produces a data
@@ -756,12 +936,17 @@ impl PE {
     /// }
     /// ```
     pub fn resolve_data_directory(&self, dir: ImageDirectoryEntry) -> Result<DataDirectory, Error> {
-        match self.get_valid_nt_headers() {
+        let directory_table = match self.get_data_directory_table() {
+            Ok(d) => d,
             Err(e) => return Err(e),
-            Ok(h) => match h {
-                NTHeaders::NTHeaders32(h32) => h32.optional_header.data_directory[dir as usize].resolve(self, dir),
-                NTHeaders::NTHeaders64(h64) => h64.optional_header.data_directory[dir as usize].resolve(self, dir),
-            }
+        };
+
+        let index = dir as usize;
+
+        if index >= directory_table.len() {
+            return Err(Error::BadDirectory);
         }
+
+        directory_table[index].resolve(self, dir)
     }
 }
