@@ -5,7 +5,7 @@ use std::slice;
 
 use widestring::U16Str;
 
-use crate::{PE, Error};
+use crate::{PE, PETranslation, Error};
 use crate::headers::*;
 
 /// Represents the architecture of the PE image.
@@ -325,7 +325,7 @@ impl Relocation {
             Err(e) => return Err(e),
         };
 
-        let offset = match self.get_address(base_rva).as_offset(pe) {
+        let offset = match pe.translate(PETranslation::Memory(self.get_address(base_rva))) {
             Ok(o) => o,
             Err(e) => return Err(e),
         };
@@ -398,6 +398,30 @@ impl Relocation {
 }
 
 /// Represents a parsed relocation entry.
+///
+/// This is ultimately the base component of the relocation table array: a base offset and some deltas.
+/// It can be used to calculate what exactly gets rewritten and where before data is modified.
+///
+/// ```rust
+/// use exe::PE;
+/// use exe::headers::ImageDirectoryEntry;
+/// use exe::types::{DataDirectory, RVA};
+///
+/// let dll = PE::from_file("test/dll.dll").unwrap();
+/// let relocation_dir = dll.resolve_data_directory(ImageDirectoryEntry::BaseReloc).unwrap();
+///
+/// if let DataDirectory::BaseReloc(relocation_table) = relocation_dir {
+///    assert_eq!(relocation_table.len(), 1);
+///
+///    let entry = &relocation_table[0];
+///    let addresses = entry.relocations
+///                         .iter()
+///                         .map(|&x| x.get_address(entry.base_relocation.virtual_address))
+///                         .collect::<Vec<RVA>>();
+///
+///    assert_eq!(addresses[0], RVA(0x1008));
+/// }
+/// ```
 pub struct RelocationEntry<'data> {
     pub base_relocation: &'data ImageBaseRelocation,
     pub relocations: &'data [Relocation]
@@ -408,7 +432,7 @@ impl<'data> RelocationEntry<'data> {
         let relocation_size = mem::size_of::<ImageBaseRelocation>();
         let word_size = mem::size_of::<u16>();
 
-        let offset = match rva.as_offset(pe) {
+        let offset = match pe.translate(PETranslation::Memory(rva)) {
             Ok(o) => o,
             Err(e) => return Err(e),
         };
@@ -469,7 +493,7 @@ pub struct RelocationEntryMut<'data> {
 impl<'data> RelocationEntryMut<'data> {
     /// Parse a mutable relocation entry at the given RVA.
     pub fn parse(pe: &'data mut PE, rva: RVA) -> Result<Self, Error> {
-        let offset = match rva.as_offset(pe) {
+        let offset = match pe.translate(PETranslation::Memory(rva)) {
             Ok(o) => o,
             Err(e) => return Err(e),
         };
@@ -525,11 +549,11 @@ impl<'data> RelocationEntryMut<'data> {
             return Err(Error::InvalidRVA);
         }
 
-        let start_offset = match start_addr.as_offset(pe) {
+        let start_offset = match pe.translate(PETranslation::Memory(start_addr)) {
             Ok(o) => o,
             Err(e) => return Err(e),
         };
-        let end_offset = match end_addr.as_offset(pe) {
+        let end_offset = match pe.translate(PETranslation::Memory(end_addr)) {
             Ok(o) => o,
             Err(e) => return Err(e),
         };
@@ -560,6 +584,26 @@ impl<'data> RelocationEntryMut<'data> {
 }
 
 /// Syntactic sugar for handling the relocation directory.
+///
+/// It can be used to quickly calculate the relocation data necessary before committing the data
+/// to memory.
+///
+/// ```rust
+/// use exe::PE;
+/// use exe::headers::ImageDirectoryEntry;
+/// use exe::types::{DataDirectory, RelocationTable, RelocationValue, RVA};
+///
+/// let dll = PE::from_file("test/dll.dll").unwrap();
+/// let relocation_dir = dll.resolve_data_directory(ImageDirectoryEntry::BaseReloc).unwrap();
+///
+/// if let DataDirectory::BaseReloc(relocation_table) = relocation_dir {
+///    let relocation_data = relocation_table.relocations(&dll, 0x02000000).unwrap();
+///    let (rva, reloc) = relocation_data[0];
+///
+///    assert_eq!(rva, RVA(0x1008));
+///    assert_eq!(reloc, RelocationValue::Relocation32(0x02001059));
+/// }
+/// ```
 pub trait RelocationTable<'data> {
     /// Get the relocation values of the given relocation table.
     fn relocations(&self, pe: &'data PE, new_base: u64) -> Result<Vec<(RVA, RelocationValue)>, Error>;
@@ -603,7 +647,7 @@ impl<'data> RelocationTable<'data> for Vec<RelocationEntry<'data>> {
         let ptr = pe.buffer.as_mut_ptr();
 
         for (rva, value) in relocations {
-            let offset = match rva.as_offset(pe) {
+            let offset = match pe.translate(PETranslation::Memory(rva)) {
                 Ok(o) => o,
                 Err(e) => return Err(e),
             };
@@ -664,7 +708,7 @@ impl<'data> RelocationTable<'data> for Vec<RelocationEntryMut<'data>> {
         let ptr = pe.buffer.as_mut_ptr();
 
         for (rva, value) in relocations {
-            let offset = match rva.as_offset(pe) {
+            let offset = match pe.translate(PETranslation::Memory(rva)) {
                 Ok(o) => o,
                 Err(e) => return Err(e),
             };
